/*
 * 2004 Viliam Holub
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "png.h"
#include "png_basics.h"

#define ERROR	-1

#define MAX_X	8192
#define MAX_Y	8192
#define HALF_X	(MAX_X/2)
#define HALF_Y	(MAX_Y/2)
#define MULT	9000
#define SCALE	16

void
die( const char *fmt, ...)

{
	va_list ap;
	
	fflush( stdout);
	
	va_start( ap, fmt);
	vfprintf( stderr, fmt, ap);
	va_end( ap);
	
	fprintf( stderr, "\n");
	exit( -1);
}

int cxl;
color_s xcxl;

void
put_3D( picture_s *pic, double ax, double ay, double az,
		double ox, double oy, double oz,
		double cx)

{
	double x, y;
	color_s c;

	c = xcxl;
	switch (cxl)
	{
		case 0:
			c.b = (127 * cx)+127;
			break;
		case 1:
			c.g = (127 * cx)+127;
			break;
		case 2:
			c.r = (127 * cx)+127;
			break;
	}

	// XXX
	az -= oz;

	if (az > 0)
	{
		x = HALF_X +ax*MULT/az;
		y = HALF_X +ay*MULT/az;
		
		put_pixel( pic, x, y, &c);
	}
}


void
clear_pic( picture_s *pic)

{
	int i;

	for (i=0; i<pic->height; i++)
		memset( pic->rows[ i], 0xff, pic->width*4);
}

#if 0
#define R1  2.879879
#define R2 -0.966918
#define R3  0.966918
#define A   0.665145
#define B   0.744728
#define C   0.744728
#endif

#if 1
#define R1  2.879879
#define R2 -0.966918
#define R3  0.966918
#define A   0.765145
#define B   0.744728
#define C   1.044728
#endif

#if 0
#define R1  1.383582e+00
#define R2 2.445375e+00
#define R3  -1.203817e+00
#define A   2.119977e+00
#define B   8.978101e-01
#define C  1.990033e+00
#endif

void
frame( double r1, double r2, double r3, double r4, double r5,
		double a, double b, double c,
		const char *fname,
		double ox, double oy, double oz)

{
	int qa, qb, wx, wy, i;
	double x, y, z, cc, xo, yo, zo;
	picture_s *pic  = new_picture();
	picture_s *pico = new_picture();

	init_picture( pic,  MAX_X, MAX_Y);
	init_picture( pico, MAX_X/SCALE, MAX_Y/SCALE);

	/*
	 * Image processing
	 */

	printf( "Creating image...\n");
	clear_pic( pic);

	printf( "Rendering...\n");
	x = y = z = cc = 1;
	for (i=0; i<4000000; i++)
	{
		put_3D( pic, x, y, z, ox, oy, oz, cc);
		xo = x;
		yo = y;
		zo = z;
		x = sin( yo*r1)+ a*sin( xo*r1);
		y = sin( xo*r2)+ b*sin( yo*r2);
		//z = 8 +sin( yo*r1);//*c;
		z = sin( yo*r4)+ c*sin( xo*r3) +3;
		cc = sin( xo*r5)+ a*sin( yo*r2);
		/*
		x = sin( yo*r1)+ a*sin( xo*r1);
		y = sin( xo*r2)+ b*sin( zo*r2);
		z = sin( zo*r3)+ c*sin( yo*r3);
		*/
	}

	/*
	 * Scale
	 */
	printf( "Scaling down...\n");
	for (qb=0; qb < MAX_X/SCALE; qb++)
		for (qa=0; qa < MAX_Y/SCALE; qa++)
		{
			color_s c = { 0, 0, 0, 0xff};

			for (wy=qb*SCALE; wy<qb*SCALE+SCALE; wy++)
				for (wx=qa*SCALE; wx<qa*SCALE+SCALE; wx++)
				{
					color_s c2;
					get_pixel( pic, &c2, wx, wy);
					c.r += c2.r;
					c.g += c2.g;
					c.b += c2.b;
				}

			c.r /= SCALE*SCALE;
			c.g /= SCALE*SCALE;
			c.b /= SCALE*SCALE;
			put_pixel( pico, qa, qb, &c);
		}

	/*
	 * Write image
	 */
	printf( "Saving picture...\n");
	save_picture( pico, fname);

	done_picture( &pic);
	done_picture( &pico);

	printf( "done\n");
}

int
main( int argc, char **argv)

{
	int i, j;
	double q[ 6], qe, qr;

	srandom( time( NULL));
	
	qe = (random()+0.0)/INT_MAX *2.0 -1;
	qr = (random()+0.0)/INT_MAX *2.0 -1;

	for (i=0; i<99; i++)
	{
		char fname[ 16];
		
		sprintf( fname, "out%02d.png", i);

		q[ 0] = R1;
		q[ 1] = R2;
		q[ 2] = R3;
		q[ 3] = A;
		q[ 4] = B;
		for (j=0; j<5; j++)
		{
			q[ j] += (random()+0.0)/INT_MAX *1.0;
		}
		for (j=5; j<6; j++)
		{
			q[ j] = (random()+0.0)/INT_MAX *2.0 +3;
			if (random() < INT_MAX/2)
				q[ j] = -q[ j];
		}
		
		qe = (random()+0.0)/INT_MAX *2.0 -1;
		qr = (random()+0.0)/INT_MAX *2.0 -1;
		
		xcxl.a = 0xff;
		xcxl.r = 0x80 +(random()&0x7f);
		xcxl.g = 0x80 +(random()&0x7f);
		xcxl.b = 0x80 +(random()&0x7f);
		cxl = random()&0x2;
		if (cxl == 0x2)
			cxl = 0;
		
		printf( "%d\t", i);
		for (j=0; j<6; j++)
			printf( "%e ", q[ j]);
		printf( "\n");
		frame( q[ 0], q[ 1], q[ 2], q[ 3], qe, qr, q[ 4], q[ 5], fname,
				0, 0, -5 /*+ (i/10.0)*1.9*/);
		q[1] += 0.01;

	}

	return 0;
}
